#pragma warning(disable:4786)

#include "DemoApp.h"

#include "boiler/Resources.h"
#include "boiler/Fluid.h"
#include "boiler/Boiler.h"
#include "utils/Random.h"

#define CAMERA_MIN_DISTANCE 190.0f
#define CAMERA_MAX_DISTANCE 400.0f
#define CAMERA_MIN_LATITUDE 0.3f
#define CAMERA_MAX_LATITUDE 3.1415927f * 0.48f


CDemoApp::CDemoApp()
{
    m_strWindowTitle    = _T("NeARAZ - Liquid Effects");
    m_bUseDepthBuffer   = TRUE;
	m_bWindowed			= TRUE;
	m_dwMinDepthBits	= 16;
	m_dwMinStencilBits	= 0;
	m_dwCreationWidth	= 640;
	m_dwCreationHeight	= 480;

	mShowingHelp = false;
	
    mFontMain		= new CD3DFont( _T("Arial"), 12, D3DFONT_BOLD );
    mFontSmall		= new CD3DFont( _T("Arial"),  9, D3DFONT_BOLD );

	mResources = new CResources();
	mObjects.addObject( *mResources );

	mFluid = new CFluid( 8 );
	mBoiler = new CBoiler( *mResources, mCamera );
	mObjects.addObject( *mBoiler );
}

CDemoApp::~CDemoApp()
{
	delete mFluid;
}


HRESULT CDemoApp::OneTimeSceneInit()
{
	ZeroMemory( mKeys, sizeof(mKeys) );

	// resources
	mResources->init( *mFluid );

	// water
	mFluid->setViscosity( 0.001f );
	mFluid->setDamping( 0.01f );
	mFluid->setTemperatureDissipation( 0.0005f );
	
    return S_OK;
}


HRESULT CDemoApp::InitDeviceObjects()
{
	resetCamera();
	
	mObjects.onCreateDevice( m_pd3dDevice );
	
    mFontMain->InitDeviceObjects( m_pd3dDevice );
    mFontSmall->InitDeviceObjects( m_pd3dDevice );

    return S_OK;
}


HRESULT CDemoApp::RestoreDeviceObjects()
{
	// camera
	mCamera.setProjectionParams(
		3.1415927f * 0.4f,
		(float)m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height,
		10.0f,
		600.0f
	);
	mCamera.updateViewParams();
	mCamera.setDeviceTransforms( *m_pd3dDevice, false, true );
	
	mObjects.onResetDevice();
	
    mFontMain->RestoreDeviceObjects();
    mFontSmall->RestoreDeviceObjects();

	return S_OK;
}


HRESULT CDemoApp::InvalidateDeviceObjects()
{
	mObjects.onLostDevice();

    mFontMain->InvalidateDeviceObjects();
    mFontSmall->InvalidateDeviceObjects();
	
    return S_OK;
}


HRESULT CDemoApp::DeleteDeviceObjects()
{
	mObjects.onResetDevice();

    mFontMain->DeleteDeviceObjects();
    mFontSmall->DeleteDeviceObjects();
	
    return S_OK;
}


// --------------------------------------------------------------------------

void CDemoApp::resetCamera()
{
	mCamera.setProjectionParams(
		3.1415927f * 0.4f,
		(float)m_d3dsdBackBuffer.Width / m_d3dsdBackBuffer.Height,
		10.0f,
		600.0f
		);
	mCamera.setDeviceTransforms( *m_pd3dDevice, false, true );

	mCameraSphericalCoords = D3DXVECTOR3( (CAMERA_MIN_DISTANCE+CAMERA_MAX_DISTANCE)/2, 0.0f, 1.0f );
	mCameraDelta = D3DXVECTOR3(0,0,0);
	mCamera.getTarget() = D3DXVECTOR3( 0.0f, 0.0f, R_FLUID_HEIGHT*0.5f );
	updateCamera();
}

void CDemoApp::updateCamera()
{
	mCameraDelta *= 0.9f;
	if( fabsf( mCameraDelta.x ) < 0.001f ) mCameraDelta.x = 0.0f;
	if( fabsf( mCameraDelta.y ) < 0.001f ) mCameraDelta.y = 0.0f;
	if( fabsf( mCameraDelta.z ) < 0.001f ) mCameraDelta.z = 0.0f;

	mCameraSphericalCoords += mCameraDelta * m_fElapsedTime;

	if( mCameraSphericalCoords.x < CAMERA_MIN_DISTANCE )
		mCameraSphericalCoords.x = CAMERA_MIN_DISTANCE;
	else if( mCameraSphericalCoords.x > CAMERA_MAX_DISTANCE )
		mCameraSphericalCoords.x = CAMERA_MAX_DISTANCE;

	if( mCameraSphericalCoords.z < CAMERA_MIN_LATITUDE )
		mCameraSphericalCoords.z = CAMERA_MIN_LATITUDE;
	else if( mCameraSphericalCoords.z > CAMERA_MAX_LATITUDE )
		mCameraSphericalCoords.z = CAMERA_MAX_LATITUDE;
	
	D3DXVECTOR3 v( mCameraSphericalCoords.x, 0.0f, 0.0f );
	D3DXMATRIX m;
	D3DXMatrixRotationY( &m, -mCameraSphericalCoords.z );
	D3DXVec3TransformCoord( &v, &v, &m );
	D3DXMatrixRotationZ( &m, mCameraSphericalCoords.y );
	D3DXVec3TransformCoord( &v, &v, &m );

	mCamera.getPosition() = v + mCamera.getTarget();
	mCamera.updateViewParams();
}

void CDemoApp::calcTemperatures( float& average, float& bottom, float& top ) const
{
	average = bottom = top = 0.0f;
	int layers = (int)(mFluid->getSize() * R_FLUID_HEIGHT / (R_FLUID_RADIUS * 2.0f));
	int cells = layers * mFluid->getLayerSize();
	int i;
	for( i = 0; i < cells; ++i ) {
		average += mFluid->getTemperature()[i];
	}
	average /= cells;
	int start = (layers-1)*mFluid->getLayerSize();
	for( i = 0; i < mFluid->getLayerSize(); ++i ) {
		bottom += mFluid->getTemperature()[i];
		top += mFluid->getTemperature()[i+start];
	}
	bottom /= mFluid->getLayerSize();
	top /= mFluid->getLayerSize();
}

HRESULT CDemoApp::FrameMove()
{
	if( mKeys['R'] ) {
		resetCamera();
	}
	// update rotation from keys
	if( mKeys[VK_UP] )		mCameraDelta.z += 0.2f;
	if( mKeys[VK_DOWN] )	mCameraDelta.z -= 0.2f;
	if( mKeys[VK_LEFT] )	mCameraDelta.y += 0.2f;
	if( mKeys[VK_RIGHT] )	mCameraDelta.y -= 0.2f;
	if( mKeys['A'] )	mCameraDelta.x -= 30.0f;
	if( mKeys['Z'] )	mCameraDelta.x += 30.0f;
	
    return S_OK;
}


HRESULT CDemoApp::Render()
{   
    HRESULT hr;
	
    if( FAILED( hr = m_pd3dDevice->BeginScene() ) )
        return hr;
	
	m_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 0x00000000, 1.0f, 0 );

	// camera
	updateCamera();
	mCamera.setDeviceTransforms( *m_pd3dDevice );

	mObjects.render( m_fElapsedTime );

	// Show info and help
	mFontSmall->DrawText( 4, 2, D3DCOLOR_ARGB(255,255,255,0), m_strFrameStats );

	if( mShowingHelp ) {
		mFontSmall->DrawText( 20, 40, D3DCOLOR_ARGB(160,255,64,64),
			"Key\n"
			"Space\n"
			"F1\n"
			"F2\n"
			"F3\n"
			"Up/Down\n"
			"Left/Right\n"
			"A/Z"
			);
		mFontSmall->DrawText( 100, 40, D3DCOLOR_ARGB(160,255,255,255),
			"action\n"
			"- heat bottom\n"
			"- toggles help\n"
			"- toggles fluid velocity display\n"
			"- toggles wireframe display\n"
			"- orbit camera vertically\n"
			"- orbit camera horizontally\n"
			"- move camera forward/backward\n"
			);
	} else {
		mFontSmall->DrawText( 20, 40, D3DCOLOR_ARGB(160,255,255,255),
			"Press F1 for help"
		);
	}
	char buf[200];
	float tavg, ttop, tbottom;
	calcTemperatures( tavg, tbottom, ttop );
	sprintf( buf, "Temperatures: avg %.0f top %.0f bottom %.0f", tavg, ttop, tbottom );
	mFontSmall->DrawText( 4, 20, D3DCOLOR_ARGB(160,255,255,255), buf );
	
	if( FAILED( hr = m_pd3dDevice->EndScene() ) )
		return hr;
		
	return S_OK;
}


LRESULT CDemoApp::MsgProc( HWND wnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
	switch( msg ) {
	case WM_KEYDOWN:
		mKeys[ wParam ] = true;
		break;
	case WM_KEYUP:
		mKeys[ wParam ] = false;
		switch( wParam ) {
		case VK_F1:
			mShowingHelp = !mShowingHelp;
			break;
		case VK_F2:
			mBoiler->setShowingVelocity( !mBoiler->isShowingVelocity() );
			break;
		case VK_F3:
			mBoiler->setDrawingWireframe( !mBoiler->isDrawingWireframe() );
			break;
		}

		break;
	case WM_CHAR:
		if( wParam == '1' ) {
			// randomize
			int sz = mFluid->getTotalSize();
			float *velX = mFluid->getVelocityX0();
			float *velY = mFluid->getVelocityY0();
			float *velZ = mFluid->getVelocityZ0();
			for( int i = 0; i < sz; ++i ) {
				*velX = random::randfs( 0.5f );
				*velY = random::randfs( 0.5f );
				*velZ = random::randfs( 0.5f );
				++velX;
				++velY;
				++velZ;
			}
		} else if( wParam == '3' ) {
			// move min. x coord
			int sz = mFluid->getSize();
			float *velX = mFluid->getVelocityX0();
			float *velY = mFluid->getVelocityY0();
			float *velZ = mFluid->getVelocityZ0();
			for( int y = 0; y < sz; ++y ) {
				for( int z = 0; z < sz; ++z ) {
					int idx = mFluid->getIndex( 0, y, z );
					mFluid->getVelocityX0()[idx] = 0.2f + random::randf(0.3f);
					mFluid->getVelocityY0()[idx] = random::randfs( 0.4f );
					mFluid->getVelocityZ0()[idx] = random::randfs( 0.4f );
				}
			}
		} else if( wParam == ' ' ) {
			// move min z coord in circle
			int sz = mFluid->getSize();
			float *velX = mFluid->getVelocityX0();
			float *velY = mFluid->getVelocityY0();
			float *velZ = mFluid->getVelocityZ0();
			float dr = 3.1415927f * 2.0f / sz / 3;
			float r = 0.0f;
			for( int i = 0; i < sz*3; ++i ) {
				int ix = sz/2 + (cosf(r)*sz/2)*0.7f;
				int iy = sz/2 + (sinf(r)*sz/2)*0.7f;
				int idx = mFluid->getIndex( ix, iy, 0 );
				//mFluid->getVelocityX0()[idx] = -sinf(r) * 0.4f;
				//mFluid->getVelocityY0()[idx] = cosf(r) * 0.4f;
				//mFluid->getVelocityZ0()[idx] = random::randf( 0.2f );
				mFluid->getTemperature0()[idx] = 100;
				r += dr;
			}
		}
		break;
	};
	
    return CD3DApplication::MsgProc( wnd, msg, wParam, lParam );
}
